import logging
import os
import re
import shutil
from datetime import datetime
from typing import Any, Dict, Generator, Optional, cast

import pytest
from _pytest.config import Config
from _pytest.config.argparsing import Parser
from _pytest.main import Session
from _pytest.python import Function
from _pytest.reports import TestReport
from _pytest.runner import CallInfo, TResult
from pluggy._result import _Result
from selenium.webdriver.chrome.webdriver import WebDriver

logging.basicConfig(level=logging.INFO)

TEMP_SCREENSHOTS_DIR_NAME = 'execution is progress...'
HISTORY_DIR_NAME = 'history'
DEFAULT_TEST_SUITE_NAME = 'UnknownTestSuite'


def pytest_addoption(parser: Parser) -> None:
    parser.addoption('--save_screenshots',
                     action='store_true',
                     default=False,
                     help='Whenever a test fails, a screenshot will immediately be taken.')
    parser.addoption('--screenshots_dir',
                     action='store',
                     default='screenshots',
                     help='''The directory where the screenshots will be saved, from project root directory.
                     By default, they are saved in the "screenshots" directory.''')


def pytest_sessionfinish(session: Session, exitstatus: int) -> None:
    """
    To avoid CI failure, forces pytest to finish its process with return code 0 when it was supposed to return 1
    See more: https://docs.pytest.org/en/7.1.x/reference/exit-codes.html
    """
    project_root_path = str(session.config.rootpath)
    screenshots_dir = session.config.getvalue('screenshots_dir')
    screenshot_dir_path = os.path.join(project_root_path, screenshots_dir)
    temp_dir_path = os.path.join(screenshot_dir_path, TEMP_SCREENSHOTS_DIR_NAME)
    last_run_dir_path = os.path.join(screenshot_dir_path, datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3])
    if os.path.exists(temp_dir_path):
        shutil.move(temp_dir_path, last_run_dir_path)


def pytest_configure(config: Config) -> None:
    if not config.getvalue('save_screenshots'):
        logging.info('Skipping screenshot: Feature is disabled by default.')
        return

    project_root_path = str(config.rootpath)
    screenshots_dir = config.getvalue('screenshots_dir')
    screenshots_root_path = os.path.join(project_root_path, screenshots_dir)

    if not os.path.exists(screenshots_root_path):
        return os.makedirs(screenshots_root_path)

    for directory in os.scandir(screenshots_root_path):
        if match_dir_date_time_format(directory.name):
            archive_old_screenshots(screenshots_root_path, directory.name)
        elif directory.name != HISTORY_DIR_NAME:
            shutil.rmtree(directory.path)


def find_web_driver_instance(test_function_arguments: Dict[str, Any]) -> Optional[WebDriver]:
    for value in test_function_arguments.values():
        if hasattr(value, 'save_screenshot'):
            return cast(WebDriver, value)
    return None


@pytest.hookimpl(hookwrapper=True)  # type: ignore
def pytest_runtest_makereport(item: Function, call: CallInfo[TResult]) -> Generator[None, None, None]:
    results_hook: _Result
    results_hook = yield
    test_result: TestReport = results_hook.get_result()
    after_test_case_routine(test_result, pytest_function=item)


def after_test_case_routine(test_result: TestReport, pytest_function: Function) -> None:
    save_screenshots = pytest_function.config.getvalue('save_screenshots')
    test_finished_phase = test_result.when == 'call'
    test_failed = test_result.failed

    if not all([save_screenshots, test_finished_phase, test_failed]):
        return

    web_driver = find_web_driver_instance(test_function_arguments=pytest_function.funcargs)
    if not web_driver:
        logging.warning('''No Selenium Web Driver instance found in any function arguments of the test context!
            Screenshots won't be taken.''')
        return

    project_root_path = str(pytest_function.config.rootpath)
    screenshots_dir = pytest_function.config.getvalue('screenshots_dir')
    session_dir_path = os.path.join(project_root_path, screenshots_dir, TEMP_SCREENSHOTS_DIR_NAME)
    test_suite_name = pytest_function.parent.name if pytest_function.parent else DEFAULT_TEST_SUITE_NAME
    capture_screenshot(web_driver, session_dir_path, test_suite=test_suite_name, test_name=pytest_function.name)


def capture_screenshot(web_driver: WebDriver, session_dir_path: str, test_suite: str, test_name: str) -> None:
    test_suite_dir_path = os.path.join(session_dir_path, test_suite)
    if not os.path.exists(test_suite_dir_path):
        os.makedirs(test_suite_dir_path)

    img_path = os.path.join(test_suite_dir_path, f'{test_name}.png')
    if os.path.exists(img_path):
        logging.warning(f'Screenshot on path "{img_path}" already exists and will be overwritten by the latest test.')

    try:
        web_driver.save_screenshot(img_path)
    except Exception as e:
        logging.error(f'Error while trying to take screen for failed {test_name=}. {e}')


def archive_old_screenshots(screenshots_root_path: str, screenshot_dir_name: str) -> None:
    dir_path_in_history = os.path.join(screenshots_root_path, HISTORY_DIR_NAME, screenshot_dir_name)
    if os.path.exists(dir_path_in_history):
        logging.warning(f'''The screenshots folder from the last execution, {dir_path_in_history}, is already archived  
        in folder in the history folder. The older screenshots will be overwritten by the newer ones.''')
        shutil.rmtree(dir_path_in_history)

    history_root_path = os.path.dirname(dir_path_in_history)
    screenshot_dir_path = os.path.join(screenshots_root_path, screenshot_dir_name)
    shutil.move(screenshot_dir_path, history_root_path)


def match_dir_date_time_format(dir_name: str) -> bool:
    date_time_regex = r'^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{3}$'
    return bool(re.search(date_time_regex, dir_name))



from pathlib import Path
from typing import Union
from unittest.mock import Mock

import pytest
from _pytest.config import Config
from _pytest.config.argparsing import Parser
from _pytest.main import Session
from _pytest.python import Function
from _pytest.reports import TestReport
from pytest_mock import MockFixture
from selenium.webdriver.chrome.webdriver import WebDriver

from src import pytest_screenshot_on_failure as psf


def get_parser_option_value(parameter: str) -> Union[bool, str]:
    if parameter == 'save_screenshots':
        return True
    elif parameter == 'screenshots_dir':
        return 'screenshot/dir/example'
    raise Exception(f'Unknown {parameter=} given.')


invalid_dir_names = [
    '2023-07-18-15:25:03.980',
    '2023-07-18.15:25:03.980',
    '2023-07-18-15:25:03.980',
    '2023-07-18:15:25:03.980',
    '2023-07-18 15:25:03-543',
    '2023.07.18 15:25:03.543',
    '2023:07:18 15:25:03.543',
    '2023-07-18 15.25.03.543',
    '2023-07-18 15-25-03.543',
    '2023.07.18 15:25:03.543',
    '2023-07-18 15.25.03.543',
    '2023-07-18 15:25:03:543',
    '2023-07-18 15:25:03',
    '2023-07-18 15:25',
    '2023-07-18',
    '15:25:03',
    'asd',
    '123',
    ''
]  # yapf: disable


@pytest.mark.parametrize('invalid_name', invalid_dir_names)
def test_dir_do_not_match_date_time_format(invalid_name: str) -> None:
    assert not psf.match_dir_date_time_format(invalid_name), f'{invalid_name=} was accepted as a valid date time format'


valid_dir_names = [
    '2023-07-18 15:25:03.980',
    '1998-12-31 20:04:58.763'
]  # yapf: disable


@pytest.mark.parametrize('valid_name', valid_dir_names)
def test_dir_match_date_time_format(valid_name: str) -> None:
    assert psf.match_dir_date_time_format(valid_name), f'{valid_name=} was not accepted as a valid date time format'


def test_archive_old_screenshots_method(mocker: MockFixture) -> None:
    mock_file_or_dir_exists = mocker.patch('os.path.exists', return_value=True)
    mock_remove_dir = mocker.patch('shutil.rmtree', return_value=True)
    mock_move_dir = mocker.patch('shutil.move', return_value=True)
    mock_log_warning = mocker.patch('logging.warning', return_value=True)

    screenshots_root_path = 'root_path/test_example'
    screenshot_dir_name = 'screenshots'
    psf.archive_old_screenshots(screenshots_root_path, screenshot_dir_name)

    mock_file_or_dir_exists.assert_called_once()
    mock_log_warning.assert_called_once()
    mock_remove_dir.assert_called_once()
    mock_move_dir.assert_called_once()


def test_capture_screenshot(mocker: MockFixture) -> None:

    def assert_image_exist(dir_name: str) -> bool:
        return '.png' in dir_name

    mock_file_or_dir_exists = mocker.patch('os.path.exists', side_effect=assert_image_exist)
    mock_create_test_suite_dir = mocker.patch('os.makedirs', return_value=True)
    mock_logging_warning = mocker.patch('logging.warning')

    mock_web_driver = Mock(spec=WebDriver)
    mock_save_screenshot = mocker.patch.object(mock_web_driver, 'save_screenshot', return_value=True)

    screenshots_root_path = 'root_path/test_example'
    psf.capture_screenshot(mock_web_driver, screenshots_root_path, test_suite='CoolTestSuite', test_name='cool_test')

    assert mock_file_or_dir_exists.call_count == 2
    mock_create_test_suite_dir.assert_called_once()
    mock_logging_warning.assert_called_once()
    mock_save_screenshot.assert_called_once()


def test_unable_to_capture_screenshot(mocker: MockFixture) -> None:

    def raise_exception(dir_path: str) -> None:
        raise Exception('Simulate exception')

    mock_file_or_dir_exists = mocker.patch('os.path.exists', return_value=True)
    mock_logging_error = mocker.patch('logging.error')

    mock_web_driver = Mock(spec=WebDriver)
    mock_save_screenshot = mocker.patch.object(mock_web_driver, 'save_screenshot', side_effect=raise_exception)

    screenshots_root_path = 'invalid/path'
    psf.capture_screenshot(mock_web_driver, screenshots_root_path, test_suite='CoolTestSuite', test_name='cool_test')

    assert mock_file_or_dir_exists.call_count == 2
    mock_save_screenshot.assert_called_once()
    mock_logging_error.assert_called_once()


def test_find_web_driver_instance() -> None:
    mock_web_driver = Mock(spec=WebDriver)

    mock_test_function_arguments = {'invalid_driver': 'web_driver', 'valid_driver': mock_web_driver}

    assert not psf.find_web_driver_instance({}), 'Expected method to return None when not finding a WebDriver instance'
    assert psf.find_web_driver_instance(mock_test_function_arguments), '''Expected method to return True when finding 
    a WebDriver instance'''


def test_skip_pytest_configure_if_there_is_no_screenshots_dir(mocker: MockFixture) -> None:
    mock_screenshot_root_exists = mocker.patch('os.path.exists', return_value=False)
    mock_make_screenshot_dir = mocker.patch('os.makedirs', return_value=True)

    mock_config = Mock(spec=Config)
    mock_config._rootpath = Path('project_root_path/example')
    mock_get_value = mocker.patch.object(mock_config, 'getvalue', side_effect=get_parser_option_value)

    psf.pytest_configure(mock_config)

    mock_make_screenshot_dir.assert_called_once()
    mock_screenshot_root_exists.assert_called_once()
    assert mock_get_value.call_count == 2


def test_skip_pytest_configure_if_feature_is_disabled(mocker: MockFixture) -> None:
    mock_logging_info = mocker.patch('logging.info')

    mock_config = Mock(spec=Config)
    mock_getvalue = mocker.patch.object(mock_config, 'getvalue', return_value=False)

    psf.pytest_configure(mock_config)

    mock_getvalue.assert_called_once()
    mock_logging_info.assert_called_once()


dir_names_list = ['history', 'not valid', '2023-07-18 15:25:03.980']
mock_dir_list = [Mock(spec=Config) for _ in range(3)]
for mock_dir, dir_name in zip(mock_dir_list, dir_names_list):
    mock_dir.name = dir_name
    mock_dir.path = 'path_to_dir'


def test_pytest_configure_archive_screenshots_and_delete_invalid_dirs(mocker: MockFixture) -> None:
    mock_screenshot_root_exists = mocker.patch('os.path.exists', return_value=True)
    mock_dir_lookup = mocker.patch('os.scandir', return_value=mock_dir_list)
    mock_archive = mocker.patch('src.pytest_screenshot_on_failure.archive_old_screenshots', return_value=True)
    mock_remove_dirs = mocker.patch('shutil.rmtree', return_value=True)

    mock_config = Mock(spec=Config)
    mock_config._rootpath = Path('project_root_path/example')
    mock_get_value = mocker.patch.object(mock_config, 'getvalue', side_effect=get_parser_option_value)

    psf.pytest_configure(mock_config)

    assert mock_get_value.call_count == 2
    mock_screenshot_root_exists.assert_called_once()
    mock_dir_lookup.assert_called_once()
    mock_archive.assert_called_once()
    mock_remove_dirs.assert_called_once()


def test_pytest_sessionfinish(mocker: MockFixture) -> None:
    mock_temp_dir_exists = mocker.patch('os.path.exists', return_value=True)
    mock_rename_dir = mocker.patch('shutil.move', return_value=True)

    mock_session = Mock(spec=Session)
    mock_session.config._rootpath = Path('project_root_path/example')
    mock_get_value = mocker.patch.object(mock_session.config, 'getvalue', side_effect=get_parser_option_value)

    psf.pytest_sessionfinish(mock_session, 0)

    mock_get_value.assert_called_once()
    mock_temp_dir_exists.assert_called_once()
    mock_rename_dir.assert_called_once()


def test_pytest_addoption(mocker: MockFixture) -> None:
    mock_parser = Mock(spec=Parser)
    mock_parser_addoption = mocker.patch.object(mock_parser, 'addoption', return_value=True)

    psf.pytest_addoption(mock_parser)

    assert mock_parser_addoption.call_count == 2


def test_conditions_to_capture_screenshot_on_test_case_routine_conditions(mocker: MockFixture) -> None:
    mock_web_driver = mocker.patch('src.pytest_screenshot_on_failure.find_web_driver_instance', return_value=None)

    mock_test_report = Mock(spec=TestReport)
    mock_test_report.when = None
    mocker.patch.object(mock_test_report, 'failed', return_value=False)
    mock_pytest_function = Mock(spec=Function)
    mock_pytest_function.funcargs = {}
    mock_getvalue = mocker.patch.object(mock_pytest_function.config, 'getvalue', return_value=False)

    psf.after_test_case_routine(mock_test_report, mock_pytest_function)

    mock_getvalue.assert_called_once()
    mock_web_driver.assert_not_called()


def test_after_test_case_routine_without_web_driver_found(mocker: MockFixture) -> None:
    mock_log_warning = mocker.patch('logging.warning', return_value=True)
    mock_find_web_driver = mocker.patch('src.pytest_screenshot_on_failure.find_web_driver_instance', return_value=None)

    mock_test_report = Mock(spec=TestReport)
    mock_test_report.when = 'call'
    mocker.patch.object(mock_test_report, 'failed', return_value=True)
    mock_pytest_function = Mock(spec=Function)
    mock_pytest_function.funcargs = {'key1': 'nothing_here', 'key2': 'nor_here'}
    mock_get_value = mocker.patch.object(mock_pytest_function.config, 'getvalue', side_effect=get_parser_option_value)

    psf.after_test_case_routine(mock_test_report, mock_pytest_function)

    mock_find_web_driver.assert_called_once()
    mock_get_value.assert_called_once()
    mock_log_warning.assert_called_once()


def test_after_test_case_routine_capturing_screenshot(mocker: MockFixture) -> None:
    mock_capture_screenshot = mocker.patch('src.pytest_screenshot_on_failure.capture_screenshot', return_value=True)
    mock_web_driver = Mock(spec=WebDriver)
    mock_web_driver_found = mocker.patch('src.pytest_screenshot_on_failure.find_web_driver_instance',
                                         return_value=mock_web_driver)

    mock_test_report = Mock(spec=TestReport)
    mock_test_report.when = 'call'
    mocker.patch.object(mock_test_report, 'failed', return_value=True)

    mock_pytest_function = Mock(spec=Function)
    mock_pytest_function.config._rootpath = Path('project_root_path/example')
    mock_pytest_function.funcargs = {'key1': 'nothing_here', 'key2': 'nor_here'}
    mock_pytest_function.name = 'Very creative test CASE name'
    mock_pytest_function.parent = Mock()
    mock_pytest_function.parent.name = 'Very creative test SUITE name'
    mock_get_value = mocker.patch.object(mock_pytest_function.config, 'getvalue', side_effect=get_parser_option_value)

    psf.after_test_case_routine(mock_test_report, mock_pytest_function)

    mock_web_driver_found.assert_called_once()
    mock_capture_screenshot.assert_called_once()
    assert mock_get_value.call_count == 2



from setuptools import find_packages, setup

with open('README.md', 'r') as readme_file:
    long_description = readme_file.read()

setup(
    name='pytest_screenshot_on_failure',
    version='1.0.0',
    description='Saves a screenshot when a test case from a pytest execution fails',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Kleber Lauton',
    url='https://github.com/kleber26/pytest-screenshot-on-failure',
    download_url='https://github.com/kleber26/pytest-screenshot-on-failure/archive/refs/tags/v1.0.0.tar.gz',
    package_dir={'': 'src'},
    py_modules=['pytest_screenshot_on_failure'],
    license='MIT',
    packages=find_packages('src'),
    keywords='pytest plugin test screenshot',
    install_requires=['pytest', 'pluggy', 'selenium', 'setuptools'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    entry_points={
        'pytest11': ['pytest_screenshot_on_failure = pytest_screenshot_on_failure',],
    },
)
